/* ReadTheDocs theme colors */


.wy-nav-top { background-color: #f07e44 }
.wy-side-nav-search { background-color: #79859E; }
.wy-nav-content { max-width: 840px }
.wy-side-nav-search input[type="text"] { border-width: 0 }

.highlight { background: rgba(50, 140, 193, 0.15); }


/* Custom classes */


.small { font-size:40% }
.smaller, .pr, .noteversion { font-size:70% }
.noteversion::after { content: "/" }


/* Code: literals and links */


.rst-content tt.literal,
.rst-content code.literal {
    color: #404040;
}
/* slim font weight for non-link code */
.rst-content tt:not(.xref),
.rst-content code:not(.xref),
.rst-content *:not(a) > tt.xref,
.rst-content *:not(a) > code.xref {
    font-weight: normal;
}


/* Just one box for annotation code for a less noisy look */


.rst-content .annotation {
    padding: 2px 5px;
    background-color: white;
    border: 1px solid #e1e4e5;
}
.rst-content .annotation tt,
.rst-content .annotation code {
    padding: 0 0;
    background-color: transparent;
    border: 0 solid transparent;
}


/* Parameter lists */


/* Mimick rubric style used for other headings */
.rst-content dl:not(.docutils) dl > dt {
    font-weight: bold;
    background: none transparent;
    border-left: none;
    margin: 0 0 12px;
    padding: 3px 0 0;
    font-size: 105%;
}

html.writer-html5 .rst-content dl[class]:not(.option-list):not(.field-list):not(.footnote):not(.glossary):not(.simple) > dt {
    color: #005A87;
    border-top: solid 3px #005A87;
    background: #8EBAE5;
}

.rst-content .viewcode-back, .rst-content .viewcode-link {
    color: #005A87;
}

/*class="longtable docutils align-default"*/
table.longtable td {
    white-space: normal!important;
}

/* Parameters contain <strong> parts and don’t need bold font */
.rst-content dl.field-list dl > dt { font-weight: unset }
/* Add colon between return tuple element name and type */
.rst-content dl:not(.docutils) dl > dt .classifier::before { content: ' : ' }
