from omnipath._core.requests.interactions._utils import import_intercell_network
from omnipath._core.requests.interactions._interactions import (
    TFmiRNA,
    Dorothea,
    OmniPath,
    TFtarget,
    KinaseExtra,
    LigRecExtra,
    PathwayExtra,
    AllInteractions,
    Transcriptional,
    PostTranslational,
    miRNA,
    lncRNAmRNA,
)
